from .polly_synthesizer import PollySynthesizer
from .xtts_synthesizer import XTTSSynthesizer
from .elevenlabs_synthesizer import ElevenlabsSynthesizer
from .openai_synthesizer import OPENAISynthesizer
from .fourie_synthesizer import FourieSynthesizer
from .deepgram_synthesizer import DeepgramSynthesizer
